package error

import (
	"embed"
	"encoding/json"
	"gitlab.schukai.com/oss/utilities/conan/header"
	"gitlab.schukai.com/oss/utilities/conan/negotiator"
	"go.uber.org/zap"
	"net/http"
	"os"
	"path/filepath"
	"reflect"
	"strconv"
	"strings"
)

type configAccess interface {
	GetResourcePath() string
	IsResourcePathSet() bool
}

func fileExists(path string) bool {
	_, err := os.Stat(path)
	return !os.IsNotExist(err)
}

type ErrorResponse struct {
	Status           int
	Hint             string
	ValidationErrors ValidationErrors
}

func NewErrorResponse(status int, hint string) ErrorResponse {
	return ErrorResponse{
		Status:           status,
		Hint:             hint,
		ValidationErrors: []ValidationError{},
	}
}

func (e ErrorResponse) Error() string {
	return "Status " + string(e.Status) + " " + e.Hint
}

func (e *ErrorResponse) AppendValidationError(err ValidationError) {
	e.ValidationErrors = append(e.ValidationErrors, err)
}
func (e *ErrorResponse) GetValidationErrors() ValidationErrors {
	return e.ValidationErrors
}

type ValidationError struct {
	Message     string
	Namespace   string
	Field       string
	StructField string
	Tag         string
	ActualTag   string
	Kind        reflect.Kind
	Type        reflect.Type
	Value       interface{}
	Param       string
}

type validationDetails struct {
	// the field affected by the review
	Message   string      `json:"message,omitempty"`
	Field     string      `json:"field,omitempty"`
	Tag       string      `json:"rules,omitempty"`
	ActualTag string      `json:"broken-rule,omitempty"`
	Value     interface{} `json:"value,omitempty"`
	Param     string      `json:"param,omitempty"`
}

type ValidationErrors []ValidationError

func StandardErrorResponse(w http.ResponseWriter, r *http.Request, errorResponse ErrorResponse) {

	cfg := r.Context().Value("config")
	emd := r.Context().Value("embed")

	neg := negotiator.New(r.Header)

	if neg.Type("text/html") != "" {

		w.Header().Set("X-Alvine-Hint", errorResponse.Hint)
		w.WriteHeader(errorResponse.Status)

		var i interface{} = cfg
		v, ok := i.(configAccess)
		if ok == true {
			if v.IsResourcePathSet() {
				wp := v.GetResourcePath()
				fp := filepath.Join(wp, "error/"+strconv.Itoa(errorResponse.Status)+".html")

				if fileExists(fp) {
					http.ServeFile(w, r, filepath.Join(wp, "error/"+strconv.Itoa(errorResponse.Status)+".html"))
					return
				}
			}
		}

		if emd != nil {
			resources := emd.(embed.FS)
			fp := filepath.Join("web/resource/error/", ""+strconv.Itoa(errorResponse.Status)+".html")

			content, err := resources.ReadFile(fp)
			if err == nil {
				w.Write(content)
				return
			}

		}
		w.Write([]byte("" + strconv.Itoa(errorResponse.Status)))

		return
	} else if neg.Type("application/json", "text/json") != "" {
		writeJsonError(w, errorResponse)
	} else {
		// https://developer.mozilla.org/en-US/docs/Web/HTTP/Status/406
		w.Header().Set("X-Alvine-Hint", errorResponse.Hint)
		w.WriteHeader(http.StatusNotAcceptable)
	}

}

type internalErrorResponseSys struct {
	Error string `json:"error,omitempty"`
	Code  int    `json:"code"`
	// a more detailed explanation of the error
	Hint string `json:"hint,omitempty"`
	// An optional field name to which this validation applies
	Validation []validationDetails `json:"validation,omitempty"`
}

type internalErrorResponse struct {
	Sys internalErrorResponseSys `json:"sys"`
}

func writeJsonError(w http.ResponseWriter, data ErrorResponse) {

	o := []validationDetails{}
	for _, a := range data.ValidationErrors {

		var rules string

		if a.Tag != a.ActualTag {
			rules = a.Tag
		}

		o = append(o, validationDetails{
			//Message: a.Tr
			Message:   a.Message,
			Field:     strings.ToLower(strings.ReplaceAll(a.Namespace, "CreateOptions.", "")),
			Tag:       rules,
			ActualTag: a.ActualTag,
			Value:     a.Value,
			Param:     a.Param,
		})
	}

	response := internalErrorResponse{
		Sys: internalErrorResponseSys{
			Error:      http.StatusText(data.Status),
			Code:       data.Status,
			Hint:       data.Hint,
			Validation: o,
		},
	}

	b, err := json.Marshal(response)
	if err != nil {
		zap.S().Error("json.Marshal fail: %s", err)
	}

	header.ApplicationJsonHeader(w)
	w.WriteHeader(data.Status)

	_, _ = w.Write(b)

}
