package header

import (
	"context"
	"database/sql/driver"
	"github.com/google/uuid"
	"net/http"
)

type RequestUUID uuid.UUID

func NewRequestUUID() RequestUUID {
	return RequestUUID(uuid.New())
}

func (r RequestUUID) UUID() uuid.UUID {
	return uuid.UUID(r)
}

func (r RequestUUID) String() string {
	return r.UUID().String()
}

// GormDataType -> sets type to binary(16)
func (r RequestUUID) GormDataType() string {
	return "binary(16)"
}

// Scan --> tells GORM how to receive from the database
func (r *RequestUUID) Scan(value interface{}) error {
	bytes, _ := value.([]byte)
	parseByte, err := uuid.FromBytes(bytes)
	*r = RequestUUID(parseByte)
	return err
}

// Value -> tells GORM how to save into the database
func (r *RequestUUID) Value() (driver.Value, error) {
	return uuid.UUID(*r).MarshalBinary()
}

// RequestIDMiddleware is a middleware that injects a request ID into the context of each
// request. A request ID is a uuid-string.
func RequestIDMiddleware(next http.Handler) http.Handler {
	fn := func(w http.ResponseWriter, r *http.Request) {
		ctx := r.Context()
		requestID := NewRequestUUID()
		w.Header().Set(XConanRequestID, requestID.String())
		ctx = context.WithValue(ctx, XConanRequestID, &requestID)
		next.ServeHTTP(w, r.WithContext(ctx))
	}
	return http.HandlerFunc(fn)
}
